" Vim ftdetect file
" Language: Snakemake (extended from python.vim)
" Maintainer: Jay Hesselberth (jay.hesselberth@gmail.com)
" Last Change: 2020 Oct 6
"
" Usage
"
" copy to $HOME/.vim/ftdetect directory
au BufNewFile,BufRead Snakefile,*.smk set filetype=snakemake 
